#pragma once
/*
 * Warren Hunt : 2006-10-09
 *
 */

#include "referencearena.h"
#include "splitstructs.h"




/*
 * Warren Hunt : 2006-10-09
 * A kd-tree frame contains all of the information relivant to building a kd-tree at a given location
 *
 */
class KDTreeFrame {
	const ObjectList &objectList;
	ReferenceArena &arena;

public:
	const float totalCost;
	const sseAABBox voxelBounds;
	const Range range;

	// constructors and = operator
	__forceinline KDTreeFrame( KDTree &input_kdTree, const ObjectList &input_objectList )
		: objectList(input_objectList), arena(input_kdTree.arena), voxelBounds(input_kdTree.worldBounds), range(0, input_objectList.size), totalCost(input_objectList.totalCost) {}
	__forceinline KDTreeFrame( KDTreeFrame &input, const sseAABBox &input_voxelBounds, const Range &input_range, const float input_totalCost )
		: objectList(input.objectList), arena(input.arena), voxelBounds(input_voxelBounds), range(input_range), totalCost(input_totalCost) {}
	__forceinline KDTreeFrame &operator =( const KDTreeFrame &input ) { new (this) KDTreeFrame(input); return *this; }


	// indexing and evaluation functions
	__forceinline BoundsPair &operator ()( const Axis axis, const unsigned __int32 index ) const { assume(index >= range.start && index < range.end && axis != NO_AXIS); return objectList[axis][arena[index]]; }
	__forceinline float cost( const unsigned __int32 index ) const { assume(index >= range.start && index < range.end); return objectList.cost(arena[index]); }
	__forceinline sseAABBox firstAABBox( void ) const { return objectList.aabb(arena[range.start]); }

	__forceinline Split defaultSplit( void ) const { return Split(totalCost); }


	/*
	 * Warren Hunt : 2006-10-09
	 * Sifts the frame into two new frames based on a given split
	 * Abstracts the kd-tree sift operation
	 *
	 */
	__forceinline Pair<KDTreeFrame, KDTreeFrame> sift( const Split &split )
	{
		assume(split.axis != NO_AXIS);
		const float plane = split.plane;
		const Array<BoundsPair> &boundsList = objectList[split.axis];

		float totalCostLower = 0.0f;
		float totalCostUpper = 0.0f;

		// sift and recurse
		const unsigned __int32 top = arena.getTop();
		arena.seek(range.start);
		if ( split.countBelow < split.countAbove )
		{
			if ( split.countBelow == 0.0f )
			{
				totalCostUpper = totalCost;
				arena.seek(range.end);
			}
			else
			{
				// appframe.end the smaller list to the end of the heap
				// compact the shorter list over the old memory
				forRange(range, cursor)
				{
					const unsigned __int32 index = arena[cursor];
					const BoundsPair bounds = boundsList[index];
					const float currentCost = objectList.cost(index);
					if ( bounds.lo < plane )
					{
						arena.append(cursor);
						totalCostLower += currentCost;
					}
					if ( bounds.hi > plane )
					{
						arena.justify(cursor);
						totalCostUpper += currentCost;
					}
				}
			}
			return makePair(KDTreeFrame(*this, voxelBounds.boundUpper(split.axis, split.plane), Range(top, arena.getTop()),              totalCostLower),
							KDTreeFrame(*this, voxelBounds.boundLower(split.axis, split.plane), Range(range.start, arena.getPosition()), totalCostUpper));
		}
		else
		{
			if ( split.countAbove == 0.0f )
			{
				totalCostLower = totalCost;
				arena.seek(range.end);
			}
			else
			{
				// appframe.end the smaller list to the end of the heap
				// compact the shorter list over the old memory
				forRange(range, cursor)
				{
					const unsigned __int32 index = arena[cursor];
					const BoundsPair bounds = boundsList[index];
					const float currentCost = objectList.cost(index);
					if ( bounds.hi > plane )
					{
						arena.append(cursor);
						totalCostUpper += currentCost;
					}
					if ( bounds.lo < plane )
					{
						arena.justify(cursor);
						totalCostLower += currentCost;
					}
				}
			}
			return makePair(KDTreeFrame(*this, voxelBounds.boundUpper(split.axis, split.plane), Range(range.start, arena.getPosition()), totalCostLower),
							KDTreeFrame(*this, voxelBounds.boundLower(split.axis, split.plane), Range(top, arena.getTop()),              totalCostUpper));
		}
	}
};
