/*
 * Warren Hunt : 2006-10-09
 *
 */

#include "stdafx.h"
#include "kdtree.h"
#include "kdframe.h"
#include "samplestructs.h"
#include "scansample.h"
#include "scanall.h"
#include "bounds.h"
#include "sort.h"



/*
 * Warren Hunt : 2006-10-09
 * The definition of a kd-tree node
 *
 */
class KDTree::KDTreeNode
{
	const unsigned __int32 flags; // holds the axis for both, the size of a leaf and the offset for the children of a node
	const float plane; // contains the offset of the first leaf object reference or the plane of a node

public:
	__forceinline KDTreeNode( void )
		: flags(NO_AXIS), plane(0.0f) {}
	__forceinline KDTreeNode( const Range &range )
		: flags(range.length()*4 + NO_AXIS), plane(reinterpret_cast(float)range.start) {}
	__forceinline KDTreeNode( const Split &split, const KDTreeNode *children )
		: flags(((unsigned __int32)(children - this))*4 + split.axis), plane(split.plane) {}

	__forceinline Axis axis( void ) const { return (Axis)(flags&0x3); }
	__forceinline bool leaf( void ) const { return axis() == NO_AXIS; }
	__forceinline unsigned __int32 childOffset( void ) const { return flags>>2; }
	__forceinline unsigned __int32 leafSize( void ) const { return flags>>2; }
	__forceinline unsigned __int32 rangeStart( void ) const { return reinterpret_cast(unsigned __int32) plane; }

	__forceinline float cost( const sseAABBox &bounds ) const
	{
		if ( leaf() )
			return (float)leafSize();
		const sseAABBox leftBounds  = bounds.boundUpper(axis(), plane);
		const sseAABBox rightBounds = bounds.boundLower(axis(), plane);
		return (leftBounds.surfaceAreaHalf()[0]*(this + childOffset())->cost(leftBounds) + rightBounds.surfaceAreaHalf()[0]*(this + childOffset() + 1)->cost(rightBounds))/bounds.surfaceAreaHalf()[0];
	}
};



/*
 * Warren Hunt : 2006-10-09
 * Split choosing abstraction (jump table)
 *
 */
__forceinline Split chooseSplit( const KDTreeFrame &frame, const SAHType type )
{
	switch ( type )
	{
	case SCANALL_ONE: return chooseSplitScanAll(frame, frame.voxelBounds.longestAxis());
	case SCANALL_ALL: return chooseSplitScanAll(frame);
	case SCANSAMPLE_ONE: return chooseSplitScanSample(frame, frame.voxelBounds.longestAxis());
	case SCANSAMPLE_ALL: return chooseSplitScanSample(frame);
	case SORT_ONE: return chooseSplitSort(frame, frame.voxelBounds.longestAxis());
	case SORT_ALL: return chooseSplitSort(frame);
	default: assume(false); return frame.defaultSplit();
	}
}



/*
 * Warren Hunt : 2006-10-09
 * The function to build a kd-tree node
 *
 */
void KDTree::buildKDTreeNode( KDTreeFrame &frame, const unsigned __int32 index )
{
	const unsigned __int32 input_count = frame.range.length();

	// build the default split cost using the quality metrics
	Split bestSplit(Sample4CostEvaluationCache::countToCostMagic(frame.totalCost)*treeQuality.leafBonus - treeQuality.traversalStepCost);

	// choose the best split according to the quality
	if ( input_count == 1 )
		bestSplit <= chooseSplitBounds(frame, frame.firstAABBox());
	else if ( input_count <= treeQuality.smallCutoff )
		bestSplit <= chooseSplitScanAll(frame);
	else if ( input_count <= treeQuality.largeCutoff )
		bestSplit <= chooseSplit(frame, treeQuality.small);
	else
		bestSplit <= chooseSplit(frame, treeQuality.large);

	// if we didn't find a good split, return
	if ( bestSplit.axis == NO_AXIS )
		new (&nodes[index]) KDTreeNode(frame.range);
	else
	{
		// sift and build the new frames
		Pair<KDTreeFrame, KDTreeFrame> childFrames = frame.sift(bestSplit);
		KDTreeFrame &frameLower = childFrames.first;
		KDTreeFrame &frameUpper = childFrames.second;

		// recurse left and right (choose the one with the best locality first)
		const unsigned __int32 childrenIndex = getNextFree();
		if ( frameLower.range.start < frameUpper.range.start )
		{
			if ( bestSplit.countBelow != 0.0f )
				buildKDTreeNode(frameLower, childrenIndex + 0);
			else
				new (&nodes[childrenIndex + 0]) KDTreeNode();
			if ( bestSplit.countAbove != 0.0f )
				buildKDTreeNode(frameUpper, childrenIndex + 1);
			else
				new (&nodes[childrenIndex + 1]) KDTreeNode();
		}
		else
		{
			if ( bestSplit.countAbove != 0.0f )
				buildKDTreeNode(frameUpper, childrenIndex + 1);
			else
				new (&nodes[childrenIndex + 1]) KDTreeNode();
			if ( bestSplit.countBelow != 0.0f )
				buildKDTreeNode(frameLower, childrenIndex + 0);
			else
				new (&nodes[childrenIndex + 0]) KDTreeNode();
		}
		new (&nodes[index]) KDTreeNode(bestSplit, &nodes[childrenIndex]);
	}
}



/*
 * Warren Hunt : 2006-10-09
 * Basic KDTree functions
 *
 */
KDTree::KDTree( const ObjectList &objectList, const KDTreeQuality &input_treeQuality )
	: worldBounds(objectList.worldBounds),
	  nodes((unsigned __int32)(2*objectList.size)),
	  arena(*(new ReferenceArena(objectList.size))),
	  treeQuality(input_treeQuality),
	  top(2)
{
	KDTreeFrame frame(*this, objectList);
	buildKDTreeNode(frame, 1);
}
KDTree::~KDTree( void )
{
	delete &arena;
}
__forceinline unsigned __int32 KDTree::getNextFree( void )
{
	top += 2;
	if ( top > nodes.size )
		nodes.resize(2*nodes.size);
	return top - 2;
}
float KDTree::cost( void ) const
{
	return nodes[1].cost(worldBounds);
}
