#pragma once
/*
 * Warren Hunt : 2006-10-09
 *
 */


#include "../basicgeometry.h"



/*
 * Warren Hunt : 2006-10-09
 * Structure containing information about a single kd-tree plane sample
 *
 */
struct Sample {
	float plane;
	float countBelow;
	float countAbove;

	__forceinline Sample( void )
		: plane(0.0f), countBelow(0.0f), countAbove(0.0f) {}
	__forceinline Sample( const float input_plane )
		: plane(input_plane), countBelow(0.0f), countAbove(0.0f) {}
	__forceinline Sample( const float input_plane, const float input_countBelow, const float input_countAbove )
		: plane(input_plane), countBelow(input_countBelow), countAbove(input_countAbove) {}
};



/*
 * Warren Hunt : 2006-10-09
 * Structure containing information about 3 kd-tree plane samples (one per axis)
 *
 */
struct Sample3 {
	sse3Point plane;
	sse3Vector countBelow;
	sse3Vector countAbove;

	__forceinline Sample3( void )
		: plane(), countBelow(), countAbove() {}
	__forceinline Sample3( const sse3Point &input_plane )
		: plane(input_plane), countBelow(), countAbove() {}
	__forceinline Sample3( const sse3Point input_plane, const sse3Vector input_countBelow, const sse3Vector input_countAbove )
		: plane(input_plane), countBelow(input_countBelow), countAbove(input_countAbove) {}
	__forceinline Sample3( const Sample &input )
		: plane(expand(input.plane)), countBelow(expand(input.countBelow)), countAbove(expand(input.countAbove)) {}
};



/*
 * Warren Hunt : 2006-10-09
 * Structure containing information about 4 kd-tree plane samples (on one axis) implemented in sse 
 *
 */
struct Sample4 {
	sse4Floats plane;
	sse4Floats countBelow;
	sse4Floats countAbove;

	__forceinline Sample4( void )
		: plane(), countBelow(), countAbove() {}
	__forceinline Sample4( const sse4Floats &input_plane )
		: plane(input_plane), countBelow(), countAbove() {}
	__forceinline Sample4( const sse4Floats input_plane, const sse4Floats input_countBelow, const sse4Floats input_countAbove )
		: plane(input_plane), countBelow(input_countBelow), countAbove(input_countAbove) {}
	__forceinline Sample4( const Sample &input )
		: plane(expand(input.plane)), countBelow(expand(input.countBelow)), countAbove(expand(input.countAbove)) {}
};


/*
 * Warren Hunt : 2006-10-09
 * Structure for calculating the cost of splits quickly
 * Abstracts the kd-tree cost function and heuristic
 * Contains code to predict the minimum cost between two sampels
 *
 */
class Sample4CostEvaluationCache {
	const sse4Floats offsetLower;
	const sse4Floats offsetUpper;
	const sse4Floats scale;

	__forceinline static float countToCostMagic( const float count ) { return countToCostMagic(expand(count))[0]; }
	__forceinline static sse4Floats countToCostMagic( const sse4Floats &counts ) { return counts; }
//	__forceinline static sse4Floats countToCostMagic( const sse4Floats &counts ) { const sse4Floats cutoff(expand(25.0f)); return blend(counts < cutoff, counts, expand(10.0f)*sqrt4(counts) - cutoff); }
	friend class Sample3CostEvaluationCache;
	friend class KDTree;

public:
	__forceinline Sample4CostEvaluationCache( const sseAABBox &voxelBounds, const Axis axis )
	{
		const sse3Vector length = voxelBounds.hi - voxelBounds.lo;
		const sse4Floats nextLength = expand(length, next(axis));
		const sse4Floats prevLength = expand(length, prev(axis));
		const sse4Floats normalizer = ~((nextLength + prevLength)*expand(length, axis) + nextLength*prevLength);

		reinterpret_cast(sse4Floats)scale = normalizer*(nextLength + prevLength);
		const sse4Floats offset = normalizer*nextLength*prevLength;

		const sse4Floats scaledLower = scale*expand(voxelBounds.lo[axis]);
		const sse4Floats scaledUpper = scale*expand(voxelBounds.hi[axis]);
		const sse4Floats correctionFactor = max4(sseZeros(), expand(2.0f) - (((offset - scaledLower) + scaledUpper) + ((offset + scaledUpper) - scaledLower)));
		reinterpret_cast(sse4Floats)offsetLower = offset + correctionFactor - scaledLower;
		reinterpret_cast(sse4Floats)offsetUpper = offset + correctionFactor + scaledUpper;
	}

	// computes the cost of a set of samples
	__forceinline sse4Floats cost( const Sample4 &samples ) const
	{
		const sse4Floats scaledPlane = scale*samples.plane;
		const sse4Floats scaledAreaLower = offsetLower + scaledPlane;
		const sse4Floats scaledAreaUpper = offsetUpper - scaledPlane;
		return scaledAreaLower*countToCostMagic(samples.countBelow) +
			   scaledAreaUpper*countToCostMagic(samples.countAbove);
	}

	// computes the cost of a set of samples using the empty space bonus
	__forceinline sse4Floats costBounds( const Sample4 &samples ) const
	{
		const sse4Floats scaledPlane = scale*samples.plane;
		const sse4Floats scaledAreaLower = offsetLower + scaledPlane;
		const sse4Floats scaledAreaUpper = offsetUpper - scaledPlane;
		return max4(scaledAreaLower, sseZeros())*scaledAreaLower*countToCostMagic(samples.countBelow) +
			   max4(scaledAreaUpper, sseZeros())*scaledAreaUpper*countToCostMagic(samples.countAbove);
	}

	// computes the range function of a set of samples (this is used in the adaptive sampling algorithm)
	__forceinline sse4Floats range( const Sample4 &samples ) const
	{
		const sse4Floats scaledPlane = scale*samples.plane;
		const sse4Floats scaledAreaLower = offsetLower + scaledPlane;
		const sse4Floats scaledAreaUpper = offsetUpper - scaledPlane;
		return scaledAreaLower*countToCostMagic(samples.countBelow) -
			   scaledAreaUpper*countToCostMagic(samples.countAbove);
	}

	// computes the minimum value between a set of samples using a quadratic approximation
	__forceinline Sample4 predictMinimum( const Sample4 &sample_0, const Sample4 &sample_1 ) const
	{
		const sse4Floats sample_0_costBelow = countToCostMagic(sample_0.countBelow);
		const sse4Floats sample_0_costAbove = countToCostMagic(sample_0.countAbove);
		const sse4Floats sample_1_costBelow = countToCostMagic(sample_1.countBelow);
		const sse4Floats sample_1_costAbove = countToCostMagic(sample_1.countAbove);

		const sse4Floats d_dPlane = ~(sample_1.plane - sample_0.plane);
		const sse4Floats dCostBelow_dPlane = (sample_1_costBelow - sample_0_costBelow)*d_dPlane;
		const sse4Floats dCostAbove_dPlane = (sample_1_costAbove - sample_0_costAbove)*d_dPlane;

		const sse4Floats scaledPlane_0 = scale*sample_0.plane;
		const sse4Floats scaledPlane_1 = scale*sample_1.plane;
		const sse4Floats dCost_0 = dCostBelow_dPlane*(offsetLower + scaledPlane_0) + dCostAbove_dPlane(offsetUpper - scaledPlane_0) + (sample_0_costBelow - sample_0_costAbove)*scale;
		const sse4Floats dCost_1 = dCostBelow_dPlane*(offsetLower + scaledPlane_1) + dCostAbove_dPlane(offsetUpper - scaledPlane_1) + (sample_1_costBelow - sample_1_costAbove)*scale;

		const sse4Floats plane = clamp4(dCost_0/((dCost_0 - dCost_1)*d_dPlane) + sample_0.plane, sample_0.plane, sample_1.plane);
		const sse4Floats weightingFactor = (plane - sample_0.plane)*d_dPlane;

		const sse4Floats countBelow = (sample_1.countBelow - sample_0.countBelow)*weightingFactor + sample_0.countBelow;
		const sse4Floats countAbove = (sample_1.countAbove - sample_0.countAbove)*weightingFactor + sample_0.countAbove;
		return Sample4(plane, countBelow, countAbove);
	}
};



/*
 * Warren Hunt : 2006-10-09
 * Structure for calculating the cost of splits quickly (one for each axis)
 * Abstracts the kd-tree cost function and heuristic
 *
 */
class Sample3CostEvaluationCache {
	const sse3Vector offsetLower;
	const sse3Vector offsetUpper;
	const sse3Vector scale;

public:
	__forceinline Sample3CostEvaluationCache( const sseAABBox &voxelBounds )
	{
		const sse3Vector length = voxelBounds.hi - voxelBounds.lo;
		const sse3Vector nextLength = rotateAxesForward(length);
		const sse3Vector prevLength = rotateAxesReverse(length);
		const sse3Vector normalizer = ~((nextLength + prevLength)(length) + (nextLength)(prevLength));

		reinterpret_cast(sse3Vector)scale = (normalizer)(nextLength + prevLength);
		const sse3Vector offset = (normalizer)(nextLength)(prevLength);

		const sse3Vector scaledLower = (scale)(voxelBounds.lo.vector());
		const sse3Vector scaledUpper = (scale)(voxelBounds.hi.vector());
		const sse3Vector correctionFactor = max(sse3Vector(), sse3Vector(expand(2.0f)) - (((offset - scaledLower) + scaledUpper) + ((offset + scaledUpper) - scaledLower)));
		reinterpret_cast(sse3Vector)offsetLower = offset + correctionFactor - scaledLower;
		reinterpret_cast(sse3Vector)offsetUpper = offset + correctionFactor + scaledUpper;
	}

	// computes the cost of a set of samples (one per axis)
	__forceinline sse3Vector cost( const Sample3 &samples ) const
	{
		const sse3Vector scaledPlane = (scale)(samples.plane.vector());
		const sse3Vector scaledAreaLower = offsetLower + scaledPlane;
		const sse3Vector scaledAreaUpper = offsetUpper - scaledPlane;
		return (scaledAreaLower)(Sample4CostEvaluationCache::countToCostMagic(reinterpret_cast(sse4Floats)samples.countBelow)) +
			   (scaledAreaUpper)(Sample4CostEvaluationCache::countToCostMagic(reinterpret_cast(sse4Floats)samples.countAbove));
	}

	// computes the cost of a set of samples (one per axis) using the empty space bonus
	__forceinline sse3Vector costBounds( const Sample3 &samples ) const
	{
		const sse3Vector scaledPlane = (scale)(samples.plane.vector());
		const sse3Vector scaledAreaLower = offsetLower + scaledPlane;
		const sse3Vector scaledAreaUpper = offsetUpper - scaledPlane;
		return max(scaledAreaLower, sse3Vector())(scaledAreaLower)(Sample4CostEvaluationCache::countToCostMagic(reinterpret_cast(sse4Floats)samples.countBelow)) +
			   max(scaledAreaUpper, sse3Vector())(scaledAreaUpper)(Sample4CostEvaluationCache::countToCostMagic(reinterpret_cast(sse4Floats)samples.countAbove));
	}
};



/*
 * Warren Hunt : 2006-10-09
 * Structure contining information about 8 kd-tree plane samples (on one axis)
 * The constructors abstract the adaptive methodology described in my paper
 *
 */
struct Sample8 {
	Sample4 lower4;
	Sample4 upper4;

	// default constructor creates a comb sample across the given axis
	__forceinline Sample8( const sseAABBox &voxelBounds, const Axis axis )
	{
		const sse4Floats bound_lo = expand(voxelBounds.lo, axis);
		const sse4Floats bound_hi = expand(voxelBounds.hi, axis);
		const sse4Floats step = (bound_hi - bound_lo)*expand(0.142858f); // 1/7 + epsilon_{machine}
		lower4.plane = bound_lo + sse4Floats(0.0f, 1.0f, 2.0f, 3.0f)*step;
		upper4.plane = lower4.plane + expand(4.0f)*step;
	}

	// adaptive constructor creates an adaptive sampling pattern where the cost function may vary widely
	__forceinline Sample8( const sseAABBox &voxelBounds, const Axis axis, const Sample8 &input, const Sample4CostEvaluationCache &costCache )
	{
		// calculate max and min values
		const sse4Floats values_lower = costCache.range(input.lower4.plane);
		const sse4Floats values_upper = costCache.range(input.upper4.plane);
		const sse4Floats totalMin = expand(values_lower, 0);
		const sse4Floats totalMax = expand(values_upper, 3);

		// check to see if the range is degenerate (avoid /0)
		if ( or4(totalMin == totalMax) )
		{
			const sse4Floats bound_lo = expand(voxelBounds.lo, axis);
			const sse4Floats bound_hi = expand(voxelBounds.hi, axis);
			const sse4Floats step = (bound_hi - bound_lo)*expand(0.111111f); // 1/9
			lower4.plane = bound_lo + sse4Floats(1.0f, 2.0f, 3.0f, 4.0f)*step;
			upper4.plane = lower4.plane + expand(4.0f)*step;
			return;
		}

		// calculates the number of ticks below each sample, for the sake of determining the number between each sample
		const sse4Floats scaler(expand(8.9999f)/(totalMax - totalMin));
		const sse4Ints indices_lower = convert((values_lower - totalMin)*scaler);
		const sse4Ints indices_upper = convert((values_upper - totalMin)*scaler);
		const sse4Ints counts_lower = shuffle<1, 2, 3 ,0>(blend(shiftLeft128<sizeof(float)>(sseOnes(indices_lower.data)), indices_lower, indices_upper)) - indices_lower;
		const sse4Ints counts_upper = shuffle<1, 2, 3 ,3>(indices_upper) - indices_upper;
		const sse4Floats step_lower = (shuffle<1, 2, 3 ,0>(blend(shiftLeft128<sizeof(float)>(sseOnes(input.lower4.plane.data)), input.lower4.plane, input.upper4.plane)) - input.lower4.plane)/convert(counts_lower + sse4Ints(1));
		const sse4Floats step_upper = (shuffle<1, 2, 3 ,3>(input.upper4.plane) - input.upper4.plane)/convert(counts_upper + sse4Ints(1));

		unsigned __int32 currentIndex = 0;
		// for the first four segments
		for ( unsigned __int32 i = 0 ; i < 4 ; i++ )
			for ( unsigned __int32 k = 1 ; k <= (unsigned __int32)counts_lower[i] ; k++, currentIndex++ )
				if ( currentIndex < 4 )
					lower4.plane[currentIndex&0x3] = input.lower4.plane[i] + step_lower[i]*k;
				else
					upper4.plane[currentIndex&0x3] = input.lower4.plane[i] + step_lower[i]*k;
		// for the last three segments
		for ( unsigned __int32 i = 0 ; i < 3 ; i++ )
			for ( unsigned __int32 k = 1 ; k <= (unsigned __int32)counts_upper[i] ; k++, currentIndex++ )
				if ( currentIndex < 4 )
					lower4.plane[currentIndex&0x3] = input.upper4.plane[i] + step_upper[i]*k;
				else
					upper4.plane[currentIndex&0x3] = input.upper4.plane[i] + step_upper[i]*k;
	}
};



