#pragma once
/*
 * Warren Hunt : 2006-10-09
 *
 */

#include "samplestructs.h"
#include "referencearena.h"
#include "bounds.h"
#include "sort.h"



/*
 * Warren Hunt : 2006-10-09
 * Chooses the best split by using a brute force (n^2) scanning algorithm on the given axis
 *
 */
__forceinline Split chooseSplitScanAll( const KDTreeFrame &frame, const Axis axis )
{
	const unsigned __int32 length = frame.range.length();
	assume(axis != NO_AXIS && length <= 36 );
	if ( veryclose(frame.voxelBounds.lo[axis], frame.voxelBounds.hi[axis]) )
		return frame.defaultSplit();

	const Sample4CostEvaluationCache costCache(frame.voxelBounds, axis);
	const Range localRange(0, length);


	// allocate a local array to scan over so we're not all over memory
	BoundsPair localArray[36];
	forRange(localRange, cursor)
		localArray[cursor] = frame(axis, frame.range.start + cursor);
	// make the array 4x in length for the sake of sse scans
	if ( (length&0x1) == 0x1 )
		localArray[length] = localArray[0];


	// initalize the best split
	Split4 bestSplits(frame.defaultSplit().cost);
	reinterpret_cast(Axis)bestSplits.axis = axis;


	// determine the overage for the unroll by 4 then scan
	const unsigned __int32 offset = (((length + (length&0x1))&0x7)>>1);
	switch ( offset )
	{
	case 0:	break;
	case 1:
		{
			Sample4 samples(load4((float *)&localArray[0]));
			forRange(localRange, cursor)
			{
				const sse4Floats object_cost(expand(frame.cost(frame.range.start + cursor)));
				const sse4Floats bounds_lo(expand(localArray[cursor].lo));
				const sse4Floats bounds_hi(expand(localArray[cursor].hi));
				samples.countBelow += object_cost&(bounds_lo < samples.plane);
				samples.countAbove += object_cost&(bounds_hi > samples.plane);
			}

			bestSplits <= Split4(samples, axis, costCache);
			break;
		}
	case 2:
		{
			Sample4 samples_0(load4((float *)&localArray[0]));
			Sample4 samples_1(load4((float *)&localArray[2]));
			forRange(localRange, cursor)
			{
				const sse4Floats object_cost(expand(frame.cost(frame.range.start + cursor)));
				const sse4Floats bounds_lo(expand(localArray[cursor].lo));
				const sse4Floats bounds_hi(expand(localArray[cursor].hi));
				samples_0.countBelow += object_cost&(bounds_lo < samples_0.plane);
				samples_0.countAbove += object_cost&(bounds_hi > samples_0.plane);
				samples_1.countBelow += object_cost&(bounds_lo < samples_1.plane);
				samples_1.countAbove += object_cost&(bounds_hi > samples_1.plane);
			}
			bestSplits <= Split4(samples_0, axis, costCache)
					   <= Split4(samples_1, axis, costCache);
			break;
		}
	case 3:
		{
			Sample4 samples_0(load4((float *)&localArray[0]));
			Sample4 samples_1(load4((float *)&localArray[2]));
			Sample4 samples_2(load4((float *)&localArray[4]));
			forRange(localRange, cursor)
			{
				const sse4Floats object_cost(expand(frame.cost(frame.range.start + cursor)));
				const sse4Floats bounds_lo(expand(localArray[cursor].lo));
				const sse4Floats bounds_hi(expand(localArray[cursor].hi));
				samples_0.countBelow += object_cost&(bounds_lo < samples_0.plane);
				samples_0.countAbove += object_cost&(bounds_hi > samples_0.plane);
				samples_1.countBelow += object_cost&(bounds_lo < samples_1.plane);
				samples_1.countAbove += object_cost&(bounds_hi > samples_1.plane);
				samples_2.countBelow += object_cost&(bounds_lo < samples_2.plane);
				samples_2.countAbove += object_cost&(bounds_hi > samples_2.plane);
			}
			bestSplits <= Split4(samples_0, axis, costCache)
					   <= Split4(samples_1, axis, costCache)
					   <= Split4(samples_2, axis, costCache);
			break;
		}
	}
	// scan
	for ( unsigned __int32 index = 2*offset ; index < length ; index += 0x8 )
	{
		Sample4 samples_0(load4((float *)&localArray[index + 0]));
		Sample4 samples_1(load4((float *)&localArray[index + 2]));
		Sample4 samples_2(load4((float *)&localArray[index + 4]));
		Sample4 samples_3(load4((float *)&localArray[index + 6]));
		forRange(localRange, cursor)
		{
			const sse4Floats object_cost(expand(frame.cost(frame.range.start + cursor)));
			const sse4Floats bounds_lo(expand(localArray[cursor].lo));
			const sse4Floats bounds_hi(expand(localArray[cursor].hi));
			samples_0.countBelow += object_cost&(bounds_lo < samples_0.plane);
			samples_0.countAbove += object_cost&(bounds_hi > samples_0.plane);
			samples_1.countBelow += object_cost&(bounds_lo < samples_1.plane);
			samples_1.countAbove += object_cost&(bounds_hi > samples_1.plane);
			samples_2.countBelow += object_cost&(bounds_lo < samples_2.plane);
			samples_2.countAbove += object_cost&(bounds_hi > samples_2.plane);
			samples_3.countBelow += object_cost&(bounds_lo < samples_3.plane);
			samples_3.countAbove += object_cost&(bounds_hi > samples_3.plane);
		}
		bestSplits <= Split4(samples_0, axis, costCache)
				   <= Split4(samples_1, axis, costCache)
				   <= Split4(samples_2, axis, costCache)
				   <= Split4(samples_3, axis, costCache);
	}

	// calculate the boundries
	sse1Float bounds_lower = localArray[0].lo;
	sse1Float bounds_upper = localArray[0].hi;
	for ( unsigned __int32 cursor = 1 ; cursor < length ; cursor++ )
	{
		bounds_lower = min1(bounds_lower, sse1Float(localArray[cursor].lo));
		bounds_upper = max1(bounds_upper, sse1Float(localArray[cursor].hi));
	}

	return bestSplits.best()
		<= chooseSplitBounds(frame, axis, costCache, BoundsPair(bounds_lower[0], bounds_upper[0]));
}



/*
 * Warren Hunt : 2006-10-09
 * Chooses the best split using a brute force (n^2) scan on all axes
 *
 */
Split chooseSplitScanAll( const KDTreeFrame &frame )
{
	return chooseSplitScanAll(frame, X_AXIS)
		<= chooseSplitScanAll(frame, Y_AXIS)
		<= chooseSplitScanAll(frame, Z_AXIS);
}
