#pragma once
/*
 * Warren Hunt : 2006-10-09
 *
 */

#include "samplestructs.h"
#include "referencearena.h"



/*
 * Warren Hunt : 2006-10-09
 * Enumeration describing the types of bounds
 *
 */
enum Bound {
	UPPER,
	LOWER
};



/*
 * Warren Hunt : 2006-10-09
 * Structure containing a split candidate (upper or lower) on an axis
 *
 */
struct SplitCandidate {
	const float plane;
	const float count;
	const Bound bound;

	__forceinline SplitCandidate( const float input_plane, const float input_count, const Bound input_bound )
		: plane(input_plane), count(input_count), bound(input_bound) {}

	__forceinline bool operator <( const SplitCandidate &input ) const { if ( plane == input.plane ) return bound < input.bound; return plane < input.plane; }
	__forceinline SplitCandidate &operator =( const SplitCandidate &input ) { new (this) SplitCandidate(input); return *this; }
};



/*
 * Warren Hunt : 2006-10-09
 * Determines the best split based on the traditional sorting method (on one axis)
 *
 */
Split chooseSplitSort( const KDTreeFrame &frame, const Axis axis )
{
	const unsigned __int32 input_count = frame.range.length();
	const Sample4CostEvaluationCache costCache(frame.voxelBounds, axis);
	const Range localRange(0, input_count);
	const Range doubleRange(0, 2*input_count);

	Split bestSplit = frame.defaultSplit();

	// build the candidate list
	float totalCount = 0.0f;
	Array<SplitCandidate> candidateList(2*input_count);
	forRange(localRange, cursor)
	{
		const unsigned __int32 index = frame.range.start + cursor;
		const float localCost = frame.cost(index);
		new (&candidateList[2*cursor + 0]) SplitCandidate(frame(axis, index).lo, localCost, LOWER);
		new (&candidateList[2*cursor + 1]) SplitCandidate(frame(axis, index).hi, localCost, UPPER);
		totalCount += localCost;
	}
	// sort the candidate list
	shellSort<SplitCandidate>(candidateList, doubleRange);


	// initalize the default sample
	Sample currentSample(0.0f, 0.0f, totalCount);

	// evaluate each split candidate in order
	forRange(doubleRange, cursor)
	{
		currentSample.plane = candidateList[cursor].plane;
		if ( candidateList[cursor].bound == LOWER )
		{
			// only check candidates that are within the current voxel
			if ( currentSample.plane > frame.voxelBounds.lo[axis] )
				bestSplit <= Split(currentSample, axis, costCache);
			currentSample.countBelow += candidateList[cursor].count;
		}
		else
		{
			currentSample.countAbove -= candidateList[cursor].count;
			// only check candidates that are within the current voxel
			if ( currentSample.plane < frame.voxelBounds.hi[axis] )
				bestSplit <= Split(currentSample, axis, costCache);
		}
	}

	return bestSplit <= chooseSplitBounds(frame, axis, costCache);
}



/*
 * Warren Hunt : 2006-10-09
 * Determines the best split based on the traditional sorting method (across all axes)
 *
 */
Split chooseSplitSort( const KDTreeFrame &frame )
{
	Split bestSplit = frame.defaultSplit();
	foreach_Axis(axis)
		bestSplit <= chooseSplitSort(frame, axis);
	return bestSplit;
}
