#pragma once
/*
 * Warren Hunt : 2006-10-09
 *
 */

#include "samplestructs.h"



/*
 * Warren Hunt : 2006-10-09
 * Contains information about a kd-tree split (candidate)
 *
 */
struct Split {
	const float plane;
	const Axis axis;
	const float cost;

	// these counts are usually approximate and only used in a performance optimization
	float countBelow;
	float countAbove;

	__forceinline Split( const float input_cost )
		: plane(0.0f), axis(NO_AXIS), cost(input_cost), countBelow(0.0f), countAbove(0.0f) {}
	__forceinline Split( const Sample &input, const Axis input_axis, const Sample4CostEvaluationCache &costCache, const bool onBounds = false )
		: plane(input.plane), axis(input_axis), cost(onBounds ? costCache.costBounds(class Sample4(input))[0] : costCache.cost(class Sample4(input))[0]), countBelow(input.countBelow), countAbove(input.countAbove) {}

	// overwrites this split with the best of this and the input split
	__forceinline Split &operator <=( const Split &input ) { if ( input.cost < cost ) new (this) Split(input); return *this; }

	void print( void ) const { printf("%c:%9f (%9f)\n", printAxis(axis), plane, cost); }
	
private:
	__forceinline Split( const float input_plane, const Axis input_axis, const float input_cost, const unsigned __int32 input_countBelow, const unsigned __int32 input_countAbove )
		: plane(input_plane), axis(input_axis), cost(input_cost), countBelow(input_countBelow), countAbove(input_countAbove) {}
	friend struct Split3;
	friend struct Split4;
};



/*
 * Warren Hunt : 2006-10-09
 * Contains information about 3 kd-tree splits (candidate) (one per axis)
 *
 */
struct Split3 {
	const sse3Point plane;
	const sse3Vector cost;

	// these counts are usually approximate and only used in a performance optimization
	sse3Vector countBelow;
	sse3Vector countAbove;

	__forceinline Split3( const float input_cost )
		: plane(), cost(expand(input_cost)), countBelow(), countAbove() {}
	__forceinline Split3( const Sample3 &input, const Sample3CostEvaluationCache &costCache, const bool onBounds = false )
		: plane(input.plane), cost(onBounds ? costCache.costBounds(input) : costCache.cost(input)), countBelow(input.countBelow), countAbove(input.countAbove) {}

	// overwrites these splits with the best of this and the input splits
	__forceinline Split3 &operator <=( const Split3 &input )
	{
		const sseMask mask = (input.cost < cost);
		reinterpret_cast(sse3Point)plane = blend(mask, input.plane, plane);
		reinterpret_cast(sse3Vector)cost = blend(mask, input.cost, cost);
		countBelow = blend(mask, input.countBelow, countBelow);
		countAbove = blend(mask, input.countAbove, countAbove);
		return *this;
	}

	// returns the best of theses splits
	__inline Split best( void ) const
	{
		const Axis bestAxis = cost[X_AXIS] < cost[Y_AXIS] ? (cost[X_AXIS] < cost[Z_AXIS] ? X_AXIS : Z_AXIS) : (cost[X_AXIS] < cost[Y_AXIS] ? X_AXIS : Y_AXIS);
		return Split(plane[bestAxis], bestAxis, cost[bestAxis], countBelow[bestAxis], countAbove[bestAxis]);
	}
};



/*
 * Warren Hunt : 2006-10-09
 * Contains information about 4 kd-tree splits (candidate)
 *
 */
struct Split4 {
	const Axis axis;
	const sse4Floats plane;
	const sse4Floats cost;

	sse4Floats countBelow;
	sse4Floats countAbove;

	__forceinline Split4( const float input_cost )
		: axis(NO_AXIS), plane(), cost(expand(input_cost)), countBelow(), countAbove() {}
	__forceinline Split4( const Sample4 &input, const Axis input_axis, const Sample4CostEvaluationCache &costCache, const bool onBounds = false )
		: plane(input.plane), axis(input_axis), cost(onBounds ? costCache.costBounds(input) : costCache.cost(input)), countBelow(input.countBelow), countAbove(input.countAbove) {}

	// overwrites these splits with the best of this and the input splits
	__forceinline Split4 &operator <=( const Split4 &input )
	{
		const sseMask mask = (input.cost < cost);
		reinterpret_cast(sse4Floats)plane = blend(mask, input.plane, plane);
		reinterpret_cast(sse4Floats)cost = blend(mask, input.cost, cost);
		countBelow = blend(mask, input.countBelow, countBelow);
		countAbove = blend(mask, input.countAbove, countAbove);
		return *this;
	}

	// returns the best of theses splits
	__inline Split best( void ) const
	{
		sse4Ints bestIndex(0, 1, 2, 3);
		sse4Floats bestCost = cost;

		sse4Ints nextIndex = shuffle<2, 3, 0, 1>(bestIndex);
		sse4Floats nextCost = shuffle<2, 3, 0, 1>(bestCost);
		sseMask mask = nextCost < bestCost;
		bestCost = blend(mask, nextCost, bestCost);
		bestIndex = blend(mask, nextIndex, bestIndex);

		nextIndex = shuffle<1, 0, 3, 2>(bestIndex);
		nextCost = shuffle<1, 0, 3, 2>(bestCost);
		mask = nextCost < bestCost;
		bestIndex = blend(mask, nextIndex, bestIndex);
		const unsigned __int32 index = bestIndex[0];

		return Split(plane[index], axis, cost[index], countBelow[index], countAbove[index]);
	}
};



/*
 * Warren Hunt : 2006-10-09
 * Assumes two sets of samples that are sorted in plane order, merges them into one sorted set
 * Uses Bitonic sort implimented in sse
 * Then attempts to find the minimum cost across the selected range of samples via projection
 *
 */
Split projectBest( const Sample8 &input_samples_0, const Sample8 &input_samples_1, const Sample4CostEvaluationCache &costCache, const Axis axis )
{
	/*
	 * Use bitonic merg to merge the samples
	 *
	 */
	sseMask mask;
	sse4Floats temp;
	Sample8 samples_0 = input_samples_0;
	Sample8 samples_1 = input_samples_1;

	// reverse samples_1
	temp = samples_1.lower4.plane;
	samples_1.lower4.plane = shuffle<3, 2, 1, 0>(samples_1.upper4.plane);
	samples_1.upper4.plane = shuffle<3, 2, 1, 0>(temp);
	temp = samples_1.lower4.countBelow;
	samples_1.lower4.countBelow = shuffle<3, 2, 1, 0>(samples_1.upper4.countBelow);
	samples_1.upper4.countBelow = shuffle<3, 2, 1, 0>(temp);
	temp = samples_1.lower4.countAbove;
	samples_1.lower4.countAbove = shuffle<3, 2, 1, 0>(samples_1.upper4.countAbove);
	samples_1.upper4.countAbove = shuffle<3, 2, 1, 0>(temp);


	// 8 wide merge
	mask = samples_0.lower4.plane < samples_1.lower4.plane;
	temp = samples_0.lower4.plane;
	samples_0.lower4.plane = blend(mask, temp, samples_1.lower4.plane);
	samples_1.lower4.plane = blend(mask, samples_1.lower4.plane, temp);
	temp = samples_0.lower4.countBelow;
	samples_0.lower4.countBelow = blend(mask, temp, samples_1.lower4.countBelow);
	samples_1.lower4.countBelow = blend(mask, samples_1.lower4.countBelow, temp);
	temp = samples_0.lower4.countAbove;
	samples_0.lower4.countAbove = blend(mask, temp, samples_1.lower4.countAbove);
	samples_1.lower4.countAbove = blend(mask, samples_1.lower4.countAbove, temp);

	mask = samples_0.upper4.plane < samples_1.upper4.plane;
	temp = samples_0.upper4.plane;
	samples_0.upper4.plane = blend(mask, temp, samples_1.upper4.plane);
	samples_1.upper4.plane = blend(mask, samples_1.upper4.plane, temp);
	temp = samples_0.upper4.countBelow;
	samples_0.upper4.countBelow = blend(mask, temp, samples_1.upper4.countBelow);
	samples_1.upper4.countBelow = blend(mask, samples_1.upper4.countBelow, temp);
	temp = samples_0.upper4.countAbove;
	samples_0.upper4.countAbove = blend(mask, temp, samples_1.upper4.countAbove);
	samples_1.upper4.countAbove = blend(mask, samples_1.upper4.countAbove, temp);


	// 4 wide merge
	mask = samples_0.lower4.plane < samples_0.upper4.plane;
	temp = samples_0.lower4.plane;
	samples_0.lower4.plane = blend(mask, temp, samples_0.upper4.plane);
	samples_0.upper4.plane = blend(mask, samples_0.upper4.plane, temp);
	temp = samples_0.lower4.countBelow;
	samples_0.lower4.countBelow = blend(mask, temp, samples_0.upper4.countBelow);
	samples_0.upper4.countBelow = blend(mask, samples_0.upper4.countBelow, temp);
	temp = samples_0.lower4.countAbove;
	samples_0.lower4.countAbove = blend(mask, temp, samples_0.upper4.countAbove);
	samples_0.upper4.countAbove = blend(mask, samples_0.upper4.countAbove, temp);

	mask = samples_1.lower4.plane < samples_1.upper4.plane;
	temp = samples_1.lower4.plane;
	samples_1.lower4.plane = blend(mask, temp, samples_1.upper4.plane);
	samples_1.upper4.plane = blend(mask, samples_1.upper4.plane, temp);
	temp = samples_1.lower4.countBelow;
	samples_1.lower4.countBelow = blend(mask, temp, samples_1.upper4.countBelow);
	samples_1.upper4.countBelow = blend(mask, samples_1.upper4.countBelow, temp);
	temp = samples_1.lower4.countAbove;
	samples_1.lower4.countAbove = blend(mask, temp, samples_1.upper4.countAbove);
	samples_1.upper4.countAbove = blend(mask, samples_1.upper4.countAbove, temp);


	// 2 and 1 wide merges
	const sseMask twowide(false, false, true, true);
	const sseMask onewide(false, true, false, true);

	temp = shuffle<2, 3, 0, 1>(samples_0.lower4.plane);
	mask = (samples_0.lower4.plane < temp) ^ twowide;
	samples_0.lower4.plane = blend(mask, samples_0.lower4.plane, temp);
	samples_0.lower4.countBelow = blend(mask, samples_0.lower4.countBelow, shuffle<2, 3, 0, 1>(samples_0.lower4.countBelow));
	samples_0.lower4.countAbove = blend(mask, samples_0.lower4.countAbove, shuffle<2, 3, 0, 1>(samples_0.lower4.countAbove));
	temp = shuffle<1, 0, 3, 2>(samples_0.lower4.plane);
	mask = (samples_0.lower4.plane < temp) ^ onewide;
	samples_0.lower4.plane = blend(mask, samples_0.lower4.plane, temp);
	samples_0.lower4.countBelow = blend(mask, samples_0.lower4.countBelow, shuffle<1, 0, 3, 2>(samples_0.lower4.countBelow));
	samples_0.lower4.countAbove = blend(mask, samples_0.lower4.countAbove, shuffle<1, 0, 3, 2>(samples_0.lower4.countAbove));

	temp = shuffle<2, 3, 0, 1>(samples_0.upper4.plane);
	mask = (samples_0.upper4.plane < temp) ^ twowide;
	samples_0.upper4.plane = blend(mask, samples_0.upper4.plane, temp);
	samples_0.upper4.countBelow = blend(mask, samples_0.upper4.countBelow, shuffle<2, 3, 0, 1>(samples_0.upper4.countBelow));
	samples_0.upper4.countAbove = blend(mask, samples_0.upper4.countAbove, shuffle<2, 3, 0, 1>(samples_0.upper4.countAbove));
	temp = shuffle<1, 0, 3, 2>(samples_0.upper4.plane);
	mask = (samples_0.upper4.plane < temp) ^ onewide;
	samples_0.upper4.plane = blend(mask, samples_0.upper4.plane, temp);
	samples_0.upper4.countBelow = blend(mask, samples_0.upper4.countBelow, shuffle<1, 0, 3, 2>(samples_0.upper4.countBelow));
	samples_0.upper4.countAbove = blend(mask, samples_0.upper4.countAbove, shuffle<1, 0, 3, 2>(samples_0.upper4.countAbove));

	temp = shuffle<2, 3, 0, 1>(samples_1.lower4.plane);
	mask = (samples_1.lower4.plane < temp) ^ twowide;
	samples_1.lower4.plane = blend(mask, samples_1.lower4.plane, temp);
	samples_1.lower4.countBelow = blend(mask, samples_1.lower4.countBelow, shuffle<2, 3, 0, 1>(samples_1.lower4.countBelow));
	samples_1.lower4.countAbove = blend(mask, samples_1.lower4.countAbove, shuffle<2, 3, 0, 1>(samples_1.lower4.countAbove));
	temp = shuffle<1, 0, 3, 2>(samples_1.lower4.plane);
	mask = (samples_1.lower4.plane < temp) ^ onewide;
	samples_1.lower4.plane = blend(mask, samples_1.lower4.plane, temp);
	samples_1.lower4.countBelow = blend(mask, samples_1.lower4.countBelow, shuffle<1, 0, 3, 2>(samples_1.lower4.countBelow));
	samples_1.lower4.countAbove = blend(mask, samples_1.lower4.countAbove, shuffle<1, 0, 3, 2>(samples_1.lower4.countAbove));

	temp = shuffle<2, 3, 0, 1>(samples_1.upper4.plane);
	mask = (samples_1.upper4.plane < temp) ^ twowide;
	samples_1.upper4.plane = blend(mask, samples_1.upper4.plane, temp);
	samples_1.upper4.countBelow = blend(mask, samples_1.upper4.countBelow, shuffle<2, 3, 0, 1>(samples_1.upper4.countBelow));
	samples_1.upper4.countAbove = blend(mask, samples_1.upper4.countAbove, shuffle<2, 3, 0, 1>(samples_1.upper4.countAbove));
	temp = shuffle<1, 0, 3, 2>(samples_1.upper4.plane);
	mask = (samples_1.upper4.plane < temp) ^ onewide;
	samples_1.upper4.plane = blend(mask, samples_1.upper4.plane, temp);
	samples_1.upper4.countBelow = blend(mask, samples_1.upper4.countBelow, shuffle<1, 0, 3, 2>(samples_1.upper4.countBelow));
	samples_1.upper4.countAbove = blend(mask, samples_1.upper4.countAbove, shuffle<1, 0, 3, 2>(samples_1.upper4.countAbove));


	/*
	 * Attempt to find the minimum cost across the selected range of samples via projection
	 *
	 */
	const sseMask upperMask = shiftLeft128<sizeof(float)>(sseOnes(onewide.data));
	return(Split4(samples_0.lower4, axis, costCache)
		<= Split4(samples_0.upper4, axis, costCache)
		<= Split4(samples_1.lower4, axis, costCache)
		<= Split4(samples_1.upper4, axis, costCache)
		<= Split4(costCache.predictMinimum(
				samples_0.lower4,
				Sample4(
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_0.lower4.plane,      samples_0.upper4.plane     )),
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_0.lower4.countBelow, samples_0.upper4.countBelow)),
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_0.lower4.countAbove, samples_0.upper4.countAbove))
				)),
				axis, costCache)
		<= Split4(costCache.predictMinimum(
				samples_0.upper4,
				Sample4(
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_0.upper4.plane,      samples_1.lower4.plane     )),
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_0.upper4.countBelow, samples_1.lower4.countBelow)),
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_0.upper4.countAbove, samples_1.lower4.countAbove))
				)),
				axis, costCache)
		<= Split4(costCache.predictMinimum(
				samples_1.lower4,
				Sample4(
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_1.lower4.plane,      samples_1.upper4.plane     )),
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_1.lower4.countBelow, samples_1.upper4.countBelow)),
					shuffle<1, 2, 3 ,0>(blend(upperMask, samples_1.lower4.countAbove, samples_1.upper4.countAbove))
				)),
				axis, costCache)
		<= Split4(costCache.predictMinimum(
				samples_1.upper4,
				Sample4(
					shuffle<1, 2, 3 ,3>(samples_1.upper4.plane     ),
					shuffle<1, 2, 3 ,3>(samples_1.upper4.countBelow),
					shuffle<1, 2, 3 ,3>(samples_1.upper4.countAbove)
				)),
				axis, costCache)
		).best();
}
